package vario.collection.mutable

import scala.collection.TraversableOnce
import scala.collection.mutable.ArrayBuffer

import vario.file.FileVariantMapper

/** Mixes [[scala.collection.mutable.ArrayBuffer]] and [[vario.file.FileVariantMapper]].
 *
 * @author myst3r10n
 */
trait HybridBufferLike[A] {

  /** Mapper of file. */
  protected val mapper: FileVariantMapper

  /** Capacity of mapper. */
  protected val capacity: Int

  /** Size of type. */
  protected val sizeOf: Int

  /** Reads from [[vario.file.FileVariantMapper]]. */
  protected def read: A

  /** Writes into [[vario.file.FileVariantMapper]]. */
  protected def write(element: A)

  /** Volatile buffer. */
  protected val buffer = ArrayBuffer[A]()

  /** Append element into hybrid buffer.
   *
   * @param element to append
   */
  def +=(element: A) = {

    buffer += element

    this

  }

  /** Append elements of other hybrid buffer into this hybrid buffer.
   *
   * @param other hybrid buffer to append
   */
  def ++=(other: HybridBufferLike[A]) = {

    other.foreach(element => buffer += element)

    this

  }

  /** Append elements of traversable object into this hybrid buffer.
   *
   * @param other traversable object to append
   */
  def ++=(xs: TraversableOnce[A]) = {

    buffer ++= xs

    this

  }

  /** Removes element from hybrid buffer.
   *
   * @param element to remove
   */
  def -=(element: A) = {

    if(mapper != null)
      throw new UnsupportedOperationException("Mapped buffer")

    buffer -= element

    this

  }

  /** Removes elements of other hybrid buffer from this hybrid buffer.
   *
   * @param other hybrid buffer to remove
   */
  def --=(other: HybridBufferLike[A]) = {

    if(mapper != null)
      throw new UnsupportedOperationException("Mapped buffer")

    other.foreach(element => buffer -= element)

    this

  }

  /** Apply element within hybrid buffer.
   *
   * @param index to apply
   */
  def apply(index: Int) =
    if(mapper == null || index > capacity)
      buffer(index)
    else {

      mapper.position = index * sizeOf
      read

    }

  /** Clear hybrid buffer. */
  def clear {

    if(mapper != null)
      throw new UnsupportedOperationException("Mapped buffer")

    buffer.clear

  }

  /** Checks whether element exists within hybrid buffer.
   *
   * @param element to check
   *
   * @return true if exists, else false
   */
  def contains(element: Any): Boolean = {

    if(mapper != null) {

      mapper.position = 0

      for(i <- 0 to capacity - 1)
        if(element == read)
          return true

    }

    buffer.contains(element)

  }

  /** Self-explanatory
   *
   * @param f self-explanatory
   */
  def foreach[U](f: A => U) = {

    if(mapper != null) {

      mapper.position = 0

      for(i <- 0 to capacity - 1)
        f(read)
    
    }

    buffer.foreach(f)

  }

  /** First element within hybrid buffer. */
  def head =
    if(mapper == null || capacity == 0)
      buffer.head
    else {

      mapper.position = 0
      read

    }

  /** Find index of element within hybrid buffer.
   *
   * @param element to find
   *
   * @return >= 0 if found, else -1
   */
  def indexOf[B >: A](element: B): Int = {

    if(mapper != null) {

      mapper.position = 0

      for(i <- 0 to capacity - 1)
        if(element == read)
          return mapper.position / sizeOf

    }

    buffer.indexOf(element)

  }

  /** Inserts element into this hybrid buffer.
   *
   * @param n position
   * @param element to insert
   */
  def insert(n: Int, element: A) {

    if(mapper != null)
      throw new UnsupportedOperationException("Mapped buffer")

    buffer.insert(n, element)

  }

  /** Inserts elements of other hybrid buffer into this hybrid buffer.
   *
   * @param n position
   * @param other hybrid buffer to insert
   */
  def insertAll(n: Int, other: HybridBufferLike[A]) {

    if(mapper != null)
      throw new UnsupportedOperationException("Mapped buffer")

    other.foreach(element => buffer.insert(n, element))

  }

  /** Last element within hybrid buffer. */
  def last =
    if(mapper == null || capacity == 0)
      buffer.last
    else {

      mapper.position = (capacity - 1) * sizeOf
      read

    }

  /** Updates element within hybrid buffer.
   *
   * @param index of element
   * @param element ot update
   */
  def update(index: Int, element: A) {

    if(mapper != null && index < capacity) {

      mapper.position = index * sizeOf
      write(element)

    } else
      buffer(index - capacity) = element

  }

  /** Number of elements within hybrid buffer. */
  def length = capacity + buffer.size

}

